package org.bidib.wizard.comm;

import java.util.BitSet;
import java.util.Calendar;
import java.util.Collection;
import java.util.List;

import org.bidib.jbidibc.Feature;
import org.bidib.jbidibc.FirmwareUpdateStat;
import org.bidib.jbidibc.Node;
import org.bidib.jbidibc.ProtocolVersion;
import org.bidib.jbidibc.SoftwareVersion;
import org.bidib.jbidibc.enumeration.FirmwareUpdateOperation;
import org.bidib.jbidibc.enumeration.IdentifyState;
import org.bidib.wizard.mvc.main.model.Accessory;
import org.bidib.wizard.mvc.main.model.AnalogPort;
import org.bidib.wizard.mvc.main.model.FeedbackPort;
import org.bidib.wizard.mvc.main.model.Flag;
import org.bidib.wizard.mvc.main.model.InputPort;
import org.bidib.wizard.mvc.main.model.LightPort;
import org.bidib.wizard.mvc.main.model.Macro;
import org.bidib.wizard.mvc.main.model.MotorPort;
import org.bidib.wizard.mvc.main.model.ServoPort;
import org.bidib.wizard.mvc.main.model.SoundPort;
import org.bidib.wizard.mvc.main.model.SwitchPort;

public interface Communication {
    /**
     * Returns the list of features of the specified node
     * 
     * @param node
     *            the node
     * @return list of features
     */
    List<Feature> getFeatures(Node node);

    void activateAnalogPort(Node node, int port, AnalogPortStatus status);

    void activateLightPort(Node node, int port, LightPortStatus status);

    void activateMotorPort(Node node, int port, MotorPortStatus status);

    void activateServoPort(Node node, int port, int value);

    void activateSoundPort(Node node, int port, SoundPortStatus status);

    void activateSwitchPort(Node node, int port, SwitchPortStatus status);

    void boosterOff(Node node);

    void boosterOn(Node node);

    void boosterQuery(Node node);

    Collection<Accessory> getAccessories(Node node);

    int getAccessoryLength(Node node);

    Collection<AnalogPort> getAnalogPorts(Node node);

    int getBoosterMaximumCurrent(Node node);

    /**
     * Get the feedback ports. If the feedback ports are not cached, a request
     * to get the number of feedback ports from the bidib node is performed.
     * 
     * @param node
     *            the node
     * @return the collection of feedback ports
     */
    Collection<FeedbackPort> getFeedbackPorts(Node node);

    /**
     * Get the feedback state, address state and confidence of the feedback
     * ports
     * 
     * @param node
     *            the node to query
     * @param begin
     *            the start range of port number
     * @param end
     *            the end range of port number
     */
    void getFeedbackPortStatus(Node node, int begin, int end);

    Collection<Flag> getFlags();

    Collection<InputPort> getInputPorts(Node node);

    void getInputPortStatus(Node node, int port);

    Collection<LightPort> getLightPorts(Node node);

    /**
     * Returns the number of entries per macro
     * 
     * @param node
     *            the node
     * @return the number of entries per macro
     */
    int getMacroLength(Node node);

    Macro getMacroContent(Node node, Macro macro);

    Collection<Macro> getMacros(Node node);

    Collection<MotorPort> getMotorPorts(Node node);

    Collection<Node> getNodes();

    ProtocolVersion getProtocolVersion(Node node);

    Collection<ServoPort> getServoPorts(Node node);

    Collection<SoundPort> getSoundPorts(Node node);

    int getStorableMacroCount(Node node);

    Collection<SwitchPort> getSwitchPorts(Node node);

    /**
     * Get the software version of the node
     * 
     * @param node
     *            the node
     * @return the current software version
     */
    SoftwareVersion getSoftwareVersion(Node node);

    void identify(Node node, IdentifyState state);

    Boolean isAddressMessagesEnabled(Node node);

    Boolean isCommandStation(Node node);

    Boolean isDccStartEnabled(Node node);

    Boolean isExternalStartEnabled(Node node);

    Boolean isFeedbackMessagesEnabled(Node node);

    Boolean isKeyMessagesEnabled(Node node);

    /**
     * Check if the node supports FW update
     * 
     * @param node
     *            the node
     * @return the node supports FW update
     */
    boolean isUpdatable(Node node);

    int readCv(Node node, int cvNumber);

    void reloadAccessory(Node node, Accessory accessory);

    void reloadMacro(Node node, Macro macro);

    void saveAccessory(Node node, Accessory accessory);

    void saveMacro(Node node, Macro macro);

    FirmwareUpdateStat sendFirmwareUpdateOperation(Node node, FirmwareUpdateOperation operation, byte... data);

    void sendTime(Calendar time, int factor);

    void setAddressMessagesEnabled(Node node, boolean enabled);

    void setBinaryState(Node node, int address, int state, boolean value);

    void setDccStartEnabled(Node node, boolean enabled);

    void setExternalStartEnabled(Node node, boolean enabled);

    void setFeedbackMessagesEnabled(Node node, boolean enabled);

    void setKeyMessagesEnabled(Node node, boolean enabled);

    void setLightPortParameters(Node node, int port, int pwmMin, int pwmMax, int dimMin, int dimMax);

    void setServoPortParameters(Node node, int port, int trimDown, int trimUp, int speed);

    void setSpeed(
        Node node, int address, SpeedSteps speedSteps, Integer speed, Direction direction, BitSet activeFunctions,
        BitSet functions);

    void startAccessory(Node node, Accessory accessory, int ascpect);

    void startMacro(Node node, Macro macro);

    void stopMacro(Node node, Macro macro);

    void sysDisable(Node node);

    void sysEnable(Node node);

    void transferMacro(Node node, Macro macro);

    void writeCv(Node node, int cvNumber, int cvValue);
}
