package org.bidib.wizard.mvc.main.model;

import java.beans.XMLDecoder;
import java.beans.XMLEncoder;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.Vector;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import org.bidib.wizard.locale.Resources;
import org.bidib.wizard.mvc.main.model.function.Function;
import org.bidib.wizard.mvc.main.model.listener.MacroListener;

public class Macro implements LabelAware {
    private final Collection<MacroListener> listeners = new LinkedList<MacroListener>();

    private int cycles = 1;

    private Vector<Function> functions = new Vector<Function>();

    private int functionSize = 0;

    private int id = 0;

    private String label = null;

    private int speed = 255;

    private Set<StartCondition> startConditions = new HashSet<StartCondition>();

    public Macro() {
    }

    public Macro(int functionSize) {
        this.functionSize = functionSize;
    }

    public void addMacroListener(MacroListener l) {
        listeners.add(l);
    }

    public void removeMacroListener(MacroListener l) {
        listeners.remove(l);
    }

    public int getCycles() {
        return cycles;
    }

    public void setCycles(int cycles) {
        this.cycles = cycles;
    }

    // equals method is overwritten, so we have to find the correct value
    // manually
    public int getFunctionIndex(Function function) {
        int result = -1;

        for (int index = 0; index < functions.size(); index++) {
            if (functions.get(index) == function) {
                result = index;
                break;
            }
        }
        return result;
    }

    private boolean addFunction(int index, Function element) {
        boolean result = false;

        if (functions.size() < functionSize) {
            this.functions.add(index, element);
            result = true;
        }
        return result;
    }

    public void addFunctionsAfter(int row, Function[] functions) {
        List<Function> functionsAdded = new LinkedList<Function>();

        if (functions != null) {
            for (int index = 0; index < functions.length; index++) {
                if (addFunction(row + index + 1, functions[index])) {
                    functionsAdded.add(functions[index]);
                }
            }
        }
        else if (addFunction(row + 1, null)) {
            functionsAdded.add(null);
        }
        if (functionsAdded.size() > 0) {
            fireFunctionsAdded(row + 1, functionsAdded.toArray(new Function[0]));
        }
    }

    public void addFunctionsBefore(int row, Function[] functions) {
        List<Function> functionsAdded = new LinkedList<Function>();

        if (functions != null) {
            for (int index = 0; index < functions.length; index++) {
                if (addFunction(row + index, functions[index])) {
                    functionsAdded.add(functions[index]);
                }
            }
        }
        else if (addFunction(row, null)) {
            functionsAdded.add(null);
        }
        if (functionsAdded.size() > 0) {
            fireFunctionsAdded(row, functionsAdded.toArray(new Function[0]));
        }
    }

    public void clearFunctions() {
        functions.clear();
        fireFunctionsRemoved();
    }

    public Function getFunction(int index) {
        Function result = null;

        if (index >= 0 && index < functions.size()) {
            result = functions.get(index);
        }
        return result;
    }

    public int getFunctionCount() {
        return functions.size();
    }

    public List<Function> getFunctions() {
        return Collections.unmodifiableList(functions);
    }

    public void setFunctions(List<Function> functions) {
        this.functions = new Vector<Function>(functions);
        if (this.functions.size() > functionSize) {
            this.functions.setSize(functionSize);
        }
    }

    public void removeFunction(int index) {
        functions.remove(index);
        fireFunctionRemoved(index);
    }

    public void replaceFunction(int index, Function function) {
        functions.set(index, function);
    }

    public int getFunctionSize() {
        return functionSize;
    }

    public void setFunctionSize(int functionSize) {
        this.functionSize = functionSize;
        if (functions.size() > functionSize) {
            functions.setSize(functionSize);
        }
    }

    public int getId() {
        return id;
    }

    public void setId(int id) {
        this.id = id;
    }

    public String getLabel() {
        return label;
    }

    public void setLabel(String label) {
        this.label = label;
        fireLabelChanged(label);
    }

    public int getSpeed() {
        return speed;
    }

    public void setSpeed(int speed) {
        this.speed = speed;
    }

    public Collection<StartCondition> getStartConditions() {
        return Collections.unmodifiableCollection(startConditions);
    }

    public void addStartCondition(StartCondition startCondition) {
        startConditions.add(startCondition);
    }

    public void removeStartCondition(StartCondition startCondition) {
        startConditions.remove(startCondition);
    }

    public void setStartConditions(Collection<StartCondition> startConditions) {
        this.startConditions = new HashSet<StartCondition>(startConditions);
    }

    private void fireFunctionsAdded(int index, Function[] functions) {
        for (MacroListener l : listeners) {
            l.functionsAdded(index, functions);
        }
    }

    private void fireFunctionRemoved(int index) {
        for (MacroListener l : listeners) {
            l.functionRemoved(index);
        }
    }

    private void fireFunctionsRemoved() {
        for (MacroListener l : listeners) {
            l.functionsRemoved();
        }
    }

    private void fireLabelChanged(String label) {
        for (MacroListener l : listeners) {
            l.labelChanged(label);
        }
    }

    public String getDebugString() {
        return getClass().getSimpleName() + "[cycles=" + cycles + ",functions=" + functions + ",functionSize="
            + functionSize + ",id=" + id + ",label=" + label + ",speed=" + speed + ",startConditions="
            + startConditions + "]";
    }

    public static Macro load(String fileName) {
        Macro result = null;
        XMLDecoder d = null;

        try {
            try {
                d =
                    new XMLDecoder(XmlLoader.changePackage(new InputStreamReader(new GZIPInputStream(
                        new FileInputStream(fileName)))));
                result = (Macro) d.readObject();
            }
            catch (IOException e) {
                d = new XMLDecoder(XmlLoader.changePackage(new InputStreamReader(new FileInputStream(fileName))));
                result = (Macro) d.readObject();
            }
        }
        catch (FileNotFoundException e) {
        }
        finally {
            if (d != null) {
                d.close();
            }
        }
        return result;
    }

    public void save(String fileName) throws IOException {
        XMLEncoder e = new XMLEncoder(new GZIPOutputStream(new BufferedOutputStream(new FileOutputStream(fileName))));

        e.writeObject(this);
        e.close();
    }

    public String toString() {
        String result = "";

        if (label != null && label.length() > 0) {
            result = label;
        }
        else {
            result = Resources.getString(getClass(), "label") + "_" + id;
        }
        return result;
    }
}
