package org.bidib.wizard.mvc.preferences.model;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Collection;
import java.util.Date;
import java.util.LinkedList;
import java.util.Properties;

import org.bidib.wizard.main.BiDiBWizard;
import org.bidib.wizard.mvc.preferences.model.listener.PreferencesListener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.qos.logback.classic.LoggerContext;
import ch.qos.logback.classic.joran.JoranConfigurator;
import ch.qos.logback.core.joran.spi.JoranException;
import ch.qos.logback.core.util.StatusPrinter;

public class Preferences {
    private static final Logger LOGGER = LoggerFactory.getLogger(Preferences.class);

    private static final String PROPERTY_COMM_PORT = "commPort";

    private static final String PROPERTY_LABEL_PATH = "labelPath";

    private static final String PROPERTY_LOG_FILE = "logFile";

    private static final String PROPERTY_RXTX_LOG_FILE = "rxtxLogFile";

    private static final String PROPERTY_MOCK = "mock";

    private static final String PROPERTY_START_TIME = "startTime";

    private static final String PROPERTY_TIME_FACTOR = "timeFactor";

    private static final String PROPERTY_WORKING_DIRECTORY = "workingDirectory";

    private static Preferences instance = null;

    private static String preferencesPath = null;

    private final Collection<PreferencesListener> listeners = new LinkedList<PreferencesListener>();

    private final SimpleDateFormat format = new SimpleDateFormat("HH:mm");

    private Properties properties;

    private Preferences() {
        properties = new Properties();
    }

    private void load() {
        FileInputStream fis = null;

        try {
            fis = new FileInputStream(new File(preferencesPath));
            properties.load(fis);
        }
        catch (FileNotFoundException e) {
            // create initial properties
            throw new IllegalArgumentException("No preferences found at: " + preferencesPath, e);
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
        finally {
            if (fis != null) {
                try {
                    fis.close();
                }
                catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }
        }
    }

    public static synchronized Preferences getInstance() {
        if (instance == null) {
            Preferences tempInstance = new Preferences();
            try {
                // try to load the properties from the file
                tempInstance.load();

                // no problem detected
                instance = tempInstance;
            }
            catch (IllegalArgumentException ex) {
                // the file is not available, create new file with default values
                tempInstance.loadDefaultValues();
                tempInstance.save();
                instance = tempInstance;
            }
        }
        return instance;
    }

    private void loadDefaultValues() {
        getLogFile();
        getRxTxLogFile();
    }

    public void addPreferencesListener(PreferencesListener l) {
        listeners.add(l);
    }

    public String getCommPort() {
        return properties.getProperty(PROPERTY_COMM_PORT);
    }

    public String getLabelPath() {
        String result = properties.getProperty(PROPERTY_LABEL_PATH);

        if (result == null || result.isEmpty()) {
            result = new File(System.getProperty("user.home"), "." + BiDiBWizard.class.getSimpleName()).getPath();
        }
        return result;
    }

    public String getLogFile() {
        String result = properties.getProperty(PROPERTY_LOG_FILE);

        if (result == null || result.isEmpty()) {
            result =
                new File(System.getProperty("java.io.tmpdir"), BiDiBWizard.class.getSimpleName() + ".log").getPath();

            // set in properties to be saved later
            properties.setProperty(PROPERTY_LOG_FILE, result);
        }
        return result;
    }

    public String getRxTxLogFile() {
        String result = properties.getProperty(PROPERTY_RXTX_LOG_FILE);

        if (result == null || result.isEmpty()) {
            result =
                new File(System.getProperty("java.io.tmpdir"), BiDiBWizard.class.getSimpleName() + "-RXTX.log")
                    .getPath();

            // set in properties to be saved later
            properties.setProperty(PROPERTY_RXTX_LOG_FILE, result);
        }
        return result;
    }

    public Date getStartTime() {
        try {
            return format.parse(properties.getProperty(PROPERTY_START_TIME, "06:00"));
        }
        catch (ParseException e) {
            throw new RuntimeException(e);
        }
    }

    public int getTimeFactor() {
        return Integer.parseInt(properties.getProperty(PROPERTY_TIME_FACTOR, "1"));
    }

    public String getWorkingDirectory() {
        return properties.getProperty(PROPERTY_WORKING_DIRECTORY);
    }

    public boolean isMock() {
        return Boolean.valueOf(properties.getProperty(PROPERTY_MOCK, "false"));
    }

    public void save() {
        FileOutputStream fos = null;

        try {
            fos = new FileOutputStream(new File(preferencesPath));
            properties.store(fos, "user properties for " + BiDiBWizard.class.getSimpleName() + " configuration");
        }
        catch (IOException e) {
            e.printStackTrace();
        }
        finally {
            if (fos != null) {
                try {
                    fos.flush(); // flush the stream
                    fos.close();
                }
                catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }
    }

    /**
     * Reload the logger configuration to make sure the changed settings for logfile location are applied.
     * Make sure the Preferences are saved before calling this method.
     */
    public void reloadLoggerConfiguration() {

        // assume SLF4J is bound to logback in the current environment
        LoggerContext context = (LoggerContext) LoggerFactory.getILoggerFactory();

        try {
            JoranConfigurator configurator = new JoranConfigurator();
            configurator.setContext(context);
            // Call context.reset() to clear any previous configuration, e.g. default 
            // configuration. For multi-step configuration, omit calling context.reset().
            context.reset();
            configurator.doConfigure(BiDiBWizard.class.getResourceAsStream("/logback.xml"));
        }
        catch (JoranException je) {
            // StatusPrinter will handle this
        }
        StatusPrinter.printInCaseOfErrorsOrWarnings(context);
    }

    public void setCommPort(String commPort) {
        properties.setProperty(PROPERTY_COMM_PORT, commPort);
        fireCommPortChanged(commPort);
    }

    public static void setPath(String path) {
        preferencesPath = path;
    }

    public void setLabelPath(String labelPath) {
        properties.setProperty(PROPERTY_LABEL_PATH, labelPath);
    }

    public void setLogFile(String logFile) {
        properties.setProperty(PROPERTY_LOG_FILE, logFile);
    }

    public void setRxTxLogFile(String rxtxLogFile) {
        properties.setProperty(PROPERTY_RXTX_LOG_FILE, rxtxLogFile);
    }

    public void setMock(boolean mock) {
        properties.setProperty(PROPERTY_MOCK, Boolean.valueOf(mock).toString());
        fireMockChanged(mock);
    }

    public void setStartTime(Date startTime) {
        properties.setProperty(PROPERTY_START_TIME, format.format(startTime));
        fireStartTimeChanged(startTime);
    }

    public void setTimeFactor(int timeFactor) {
        properties.setProperty(PROPERTY_TIME_FACTOR, String.valueOf(timeFactor));
        fireTimeFactorChanged(timeFactor);
    }

    public void setWorkingDirectory(String workingDirectory) {
        properties.setProperty(PROPERTY_WORKING_DIRECTORY, workingDirectory);
    }

    private void fireCommPortChanged(String commPort) {
        for (PreferencesListener l : listeners) {
            LOGGER.debug("Notify listener that the comm port has changed: {}", l);
            l.commPortChanged(commPort);
        }
    }

    private void fireMockChanged(boolean mock) {
        for (PreferencesListener l : listeners) {
            l.mockChanged(mock);
        }
    }

    private void fireStartTimeChanged(Date startTime) {
        for (PreferencesListener l : listeners) {
            l.startTimeChanged(startTime);
        }
    }

    private void fireTimeFactorChanged(int timeFactor) {
        for (PreferencesListener l : listeners) {
            l.timeFactorChanged(timeFactor);
        }
    }
}
